/* 
   ORSA - Orbit Reconstruction, Simulation and Analysis
   Copyright (C) 2002-2004 Pasquale Tricarico
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   As a special exception, Pasquale Tricarico gives permission to
   link this program with Qt commercial edition, and distribute the
   resulting executable, without including the source code for the Qt
   commercial edition in the source distribution.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifndef _XORSA_OPENGL_H_
#define _XORSA_OPENGL_H_

// this header must go first, otherwise bcc55 will complain loudly
#include <qgl.h>

#include <string>
#include <map>

// GL2PS
#include "gl2ps.h"

#include "glcontrolwidget.h"

#include <qmainwindow.h>
#include <qtabwidget.h> 
#include <qcombobox.h>
#include <qfont.h>
#include <qapplication.h>
#include <qtooltip.h> 
#include <qgroupbox.h>
#include <qdatetime.h> 
#include <qtimer.h>
#include <qslider.h>
#include <qtoolbutton.h> 

#include <orsa_universe.h>
#include <orsa_orbit.h>

#include "xorsa_units_combo.h"
#include "xorsa_location_selector.h"
#include "xorsa_objects_combo.h"

class QVBoxLayout;
class QPopupMenu;
class QPushButton;
class QLabel;
class QDockWindow;
class QVBox;
class QRadioButton;
class QCheckBox;

// icons
#include "play.xpm"
// #include "l4.xpm"
#include "lagrange.xpm"
#include "camera_position.xpm"
#include "rotation.xpm"
//
// #include "tmp.xpm"

GLboolean invert(GLdouble src[16], GLdouble inverse[16]);

enum OpenGL_Labels_Mode { MASSLESS,
			  MASSIVE,
			  ALL };

// check gl2ps.h
enum OpenGL_Export_File { OGL_PS_FILE  = GL2PS_PS,
			  OGL_EPS_FILE = GL2PS_EPS,
			  OGL_PDF_FILE = GL2PS_PDF };

enum OpenGL_Projection { OGL_ORTHO,
			 OGL_PERSPECTIVE };

class XOrsaLabelsModeCombo : public QComboBox {
  
  Q_OBJECT
  
 public:
  XOrsaLabelsModeCombo(QWidget *parent=0);
  
 private slots:
  void SetLabelsMode(int);
  
 public slots:
  void SetLabelsMode(OpenGL_Labels_Mode);
  
 signals:
  void ModeChanged(OpenGL_Labels_Mode);
  
 public:  
  OpenGL_Labels_Mode GetLabelsMode();
  
 private:
  OpenGL_Labels_Mode m; 
};

class OrbitCache {
 public:
  OrbitCache() {
    set=false;
  }
  orsa::Orbit orbit;
  int ref_body_index;
  bool set;
};

// 'enhanced' types

class BoolObject : public QObject {
  
  Q_OBJECT
  
 public:
  BoolObject() : QObject() { 
    _b = false;
  }
  
  BoolObject(const bool b) : QObject() { 
    _b = b;
  }
  
 public:
  bool operator = (const bool b) {
    if (b != _b) {
      _b = b;
      emit changed();
    }
    return _b;
  }
  
  operator bool() const {
    return _b;
  }
  
 signals:
  void changed();
  
 private:
  bool _b;  
};

class IntObject : public QObject {
  
  Q_OBJECT
  
 public:
  IntObject() : QObject() { 
    _i = 0;
  }
    
  IntObject(const int i) : QObject() { 
    _i = i;
  }
  
 public:
  int operator = (const int i) {
    if (i != _i) {
      _i = i;
      emit changed();
    }
    return _i;
  }
  
  int operator = (const IntObject &int_obj) {
    if (int_obj != _i) {
      _i = int_obj;
      emit changed();
    }
    return _i;
  }
  
  operator int() const {
    return _i;
  }	
  
  // ++ and -- (prefix) operators 
  
 public:
  IntObject & operator ++ () {
    ++_i;
    emit changed();
    return *this;
  }
  
  IntObject & operator -- () {
    --_i;
    emit changed();
    return *this;
  }
  
 signals:
  void changed();
  
 private:
  int _i;
};

// periodic: from 0 to size-1;
// if equal to size-1, the ++ operator sets it to 0;
// if equal to 0, the -- operator sets it to size-1;
// useful to save a position inside a vector;
class SizeObject : public QObject {
  
  Q_OBJECT
  
 public:
  SizeObject() : QObject() { 
    internal_change = false;     
    _i = 0;
    _s = 1;
  }
  
 public:
  unsigned int operator = (const unsigned int i) {
    if (!internal_change) {
      internal_change = true;
      //
      if (i != _i) {
	_i = i % _s;
	emit changed();
      }
      //
      internal_change = false;
    }
    return _i;
  }
  
  operator unsigned int () const {
    return _i;
  }
  
  void SetSize(unsigned int s) {
    if (!internal_change) {
      internal_change = true;
      //
      if (s != _s) {
	_s = s;
	emit size_changed();
	_i = _i % _s;
      }
      //
      internal_change = false;
    }
  }
  
  unsigned int Size() const {
    return _s;
  }
  
  // ++ and -- (prefix) operators 
  
 public:
  SizeObject & operator ++ () {
    if (!internal_change) {
      internal_change = true;
      //
      _i = ++_i % _s;
      emit changed();
      //
      internal_change = false;
    }      
    return *this;
  }
  
  SizeObject & operator -- () {
    if (!internal_change) {
      internal_change = true;
      //
      _i = (_s + --_i) % _s;
      emit changed();
      //
      internal_change = false;
    }  
    return *this;
  }
  
 private:
  bool internal_change;
 signals:
  void changed();
  void size_changed();
 private:
  unsigned int _i;
 private:
  unsigned int _s;
};

class DoubleObject : public QObject {
  
  Q_OBJECT
  
 public:
  DoubleObject() : QObject() { 
    _d = 0.0;
  }
  
  DoubleObject(const double d) : QObject() { 
    _d = d;
  }
  
 public:
  double operator = (const double d) {
    if (d != _d) {
      _d = d;
      emit changed();
    }
    return _d;
  }
  
  operator double() const {
    return _d;
  }	
  
  DoubleObject & operator += (const DoubleObject & d) {
    if (d != 0.0) {
      _d += d;
      emit changed();
    }
    return *this;
  }
  
  DoubleObject & operator -= (const DoubleObject & d) {
    if (d != 0.0) {
      _d -= d;
      emit changed();
    }
    return *this;
  }
  
  DoubleObject & operator *= (const DoubleObject & d) {
    if (d != 1.0) {
      _d *= d;
      emit changed();
    }
    return *this;
  }
  
 signals:
  void changed();
  
 private:
  double _d;
};

class DoubleObjectWithLimits : public QObject {
  
  Q_OBJECT
  
 public:
  DoubleObjectWithLimits() : QObject() { 
    _d = _min = _max = 0.0;
  }
  
  DoubleObjectWithLimits(const double d) : QObject() { 
    _d = _min = _max = d;
  }
  
 public:
  double operator = (const double d) {
    if (d != _d) {
      _d = d;
      check_limits();
      emit changed();
    }
    return _d;
  }
  
  operator double() const {
    return _d;
  }
  
  void SetMin(double m) {
    if (m != _min) {
      _min = MIN(m,_max);
      _max = MAX(m,_max);
      check_limits();
      emit limits_changed();
    }
  }
  
  void SetMax(double M) {
    if (M != _max) {
      _min = MIN(M,_min);
      _max = MAX(M,_min);
      check_limits();
      emit limits_changed();
    }
  }
  
  void SetMinMax(double m, double M) {
    if ((m != _min) || (M != _max)) {
      _min = MIN(m,M);
      _max = MAX(m,M);
      check_limits();
      emit limits_changed();
    }
  }
  
  double min() const {
    return _min;
  }
  
  double max() const {
    return _max;
  }
  
  DoubleObjectWithLimits & operator += (const DoubleObjectWithLimits & d) {
    if (d != 0.0) {
      _d += d;
      check_limits();
      emit changed();
    }
    return *this;
  }
  
  DoubleObjectWithLimits & operator -= (const DoubleObjectWithLimits & d) {
    if (d != 0.0) {
      _d -= d;
      check_limits();
      emit changed();
    }
    return *this;
  }
  
  DoubleObjectWithLimits & operator *= (const DoubleObjectWithLimits & d) {
    if (d != 1.0) {
      _d *= d;
      check_limits();
      emit changed();
    }
    return *this;
  }
  
 signals:
  void changed();
  void limits_changed();
  
 private:
  double _d;
  
 private:
  double _min, _max;
  
 private:
  void check_limits() {
    if (_d < _min) {
      {
	char err_msg[4096];
	sprintf(err_msg,"DoubleObjectWithLimits::check_limits(): %g < %g",_d,_min);
	ORSA_ERROR_MACRO(err_msg);
      }
      _d = _min;
      emit changed();
    } 
    if (_d > _max) {
      {
	char err_msg[4096];
	sprintf(err_msg,"DoubleObjectWithLimits::check_limits(): %g > %g",_d,_max);
	ORSA_ERROR_MACRO(err_msg);
      }
      _d = _max;
      emit changed();
    }
  }
};


class DoubleObjectPeriodic : public QObject {
  
  Q_OBJECT
  
 public:
  DoubleObjectPeriodic() : QObject() { 
    _d = _min = _max = 0.0;
    locked = false;
  }
  
  DoubleObjectPeriodic(const double d) : QObject() { 
    _d = _min = _max = d;
    locked = false;
  }
  
 public:
  double operator = (const double d) {
    if (!locked && (d != _d)) {
      _d = d;
      check_limits();
      emit changed();
    }
    return _d;
  }
  
  operator double() const {
    return _d;
  }
  
  void SetMin(double m) {
    if (locked) {
      ORSA_ERROR_MACRO("DoubleObjectPeriodic::SetMin(): Warning: called SetMin() for a locked variable...");
    }
    if (m != _min) {
      _min = MIN(m,_max);
      _max = MAX(m,_max);
      check_limits();
      emit limits_changed();
    }
  }
  
  void SetMax(double M) {
    if (locked) {
      ORSA_ERROR_MACRO("DoubleObjectPeriodic::SetMax(): Warning: called SetMax() for a locked variable...");
    }
    if (M != _max) {
      _min = MIN(M,_min);
      _max = MAX(M,_min);
      check_limits();
      emit limits_changed();
    }
  }
  
  void SetMinMax(double m, double M) {
    if (locked) {
      ORSA_ERROR_MACRO("DoubleObjectPeriodic::SetMinMax(): Warning: called SetMinMax() for a locked variable...");
    }
    if ((m != _min) || (M != _max)) {
      _min = MIN(m,M);
      _max = MAX(m,M);
      check_limits();
      emit limits_changed();
    }
  }
  
  double min() const {
    return _min;
  }
  
  double max() const {
    return _max;
  }
  
  void Lock(const double d) {
    if (locked && (d != _d)) {
      ORSA_ERROR_MACRO("DoubleObjectPeriodic::Lock(): Warning: this variable was already locked to a different value...");
    }	
    _d = d;
    locked = true;
    emit changed();
  }
  
  void UnLock() {
    if (!locked) {
      ORSA_ERROR_MACRO("DoubleObjectPeriodic::UnLock(): Warning: this variable was not locked...");
    }
    locked = false;
  }
  
  bool Locked() const {
    return locked;
  }
  
  DoubleObjectPeriodic & operator += (const DoubleObjectPeriodic & d) {
    if (!locked && (d != 0.0)) {
      _d += d;
      check_limits();
      emit changed();
    }
    return *this;
  }
  
  DoubleObjectPeriodic & operator -= (const DoubleObjectPeriodic & d) {
    if (!locked && (d != 0.0)) {
      _d -= d;
      check_limits();
      emit changed();
    }
    return *this;
  }
  
  DoubleObjectPeriodic & operator *= (const DoubleObjectPeriodic & d) {
    if (!locked && (d != 1.0)) {
      _d *= d;
      check_limits();
      emit changed();
    }
    return *this;
  }
  
 signals:
  void changed();
  void limits_changed();
  
 private:
  double _d;
  
 private:
  double _min, _max;
  
 private:
  bool locked;
  
 private:
  double periodic_mod(const double x, const double min, const double max) {
    const double period = max - min;
    return (min+fmod(fmod(x-min,period)+period,period));
  }
  
 private:
  void check_limits() {
    if (locked) return;
    if (_d < _min) {
      _d = periodic_mod(_d,_min,_max);
      emit changed();
    } 
    if (_d > _max) {
      _d = periodic_mod(_d,_min,_max);
      emit changed();
    }
  }
};

class VectorObject : public QObject {
  
  Q_OBJECT
  
 public:
  VectorObject() : QObject() { 
    _v.Set(0,0,0);
  }

  VectorObject(const orsa::Vector &v) : QObject() { 
    _v = v;
  }
  
 public:
  orsa::Vector & operator = (const orsa::Vector &v) {
    if (v != _v) {
      _v = v;
      emit changed();
    }
    return _v;
  }
  
  operator orsa::Vector() const {
    return _v;
  }	
  
 signals:
  void changed();
  
 private:
  orsa::Vector _v;  
};

// special ones
class ProjectionObject : public QObject {
  
  Q_OBJECT
  
 public:
  ProjectionObject() : QObject() { 
    _p = OGL_ORTHO;
  }
  
  ProjectionObject(const OpenGL_Projection p) : QObject() { 
    _p = p;
  }
  
 public:
  OpenGL_Projection operator = (const OpenGL_Projection p) {
    if (p != _p) {
      _p = p;
      emit changed();
    }
    return _p;
  }
  
  operator OpenGL_Projection() const {
    return _p;
  }	
  
 signals:
  void changed();
  
 private:
  OpenGL_Projection _p;  
};

class LabelsModeObject : public QObject {
  
  Q_OBJECT
  
 public:
  LabelsModeObject() : QObject() { 
    _m = MASSIVE;
  }
  
  LabelsModeObject(const OpenGL_Labels_Mode m) : QObject() { 
    _m = m;
  }
  
 public:
  OpenGL_Labels_Mode operator = (const OpenGL_Labels_Mode m) {
    if (m != _m) {
      _m = m;
      emit changed();
    }
    return _m;
  }
  
  operator OpenGL_Labels_Mode() const {
    return _m;
  }	
  
 signals:
  void changed();
  
 private:
  OpenGL_Labels_Mode _m;  
};



class FPS {
 public:
  FPS() : max_vec_size(64) {
    fps = 0.0;
    qtime.start();
    fps_running_average_index = 0;    
  }
  void sample() {
    const double sample_fps = 1000.0/qtime.elapsed();
    if (fps_running_average_vector.size() < max_vec_size) {
      fps_running_average_vector.push_back(sample_fps);
      fps = (fps*(fps_running_average_vector.size()-1)+sample_fps)/(fps_running_average_vector.size());
    } else {
      const double old_sample_fps = fps_running_average_vector[fps_running_average_index];
      fps_running_average_vector[fps_running_average_index] = sample_fps;
      fps += (sample_fps-old_sample_fps)/max_vec_size;
      fps_running_average_index = (fps_running_average_index+1)%(fps_running_average_vector.size());
    }
    qtime.restart();
  }
  
  operator double () const {
    return fps;
  }
  
  void reset() {
    fps = 0.0;
    fps_running_average_vector.clear();
    fps_running_average_index = 0;
    qtime.restart();
  }
 private:
  double  fps;
  QTime qtime;
  const unsigned int  max_vec_size;
  std::vector<double> fps_running_average_vector;
  unsigned int        fps_running_average_index;
};

// Tools

class XOrsaBoolToolButton : public QToolButton {
  
  Q_OBJECT
    
 public:
  XOrsaBoolToolButton(BoolObject *bool_obj, const QIconSet & iconSet, const QString & textLabel, QToolBar * parent, bool toggle_button=true) : QToolButton(iconSet,textLabel,QString::null,0,0,parent), obj(bool_obj) { 
    setToggleButton(toggle_button);
    internal_change = false;
    slot_object_changed();
    connect(obj,SIGNAL(changed()),this,SLOT(slot_object_changed()));
    connect(this,SIGNAL(toggled(bool)),this,SLOT(slot_button_toggled()));
  }
  
 private slots:
  void slot_object_changed() {
    if (!internal_change) {
      internal_change = true;
      //
      if (*obj) {
	setState(QButton::On);
      } else {
	setState(QButton::Off);
      }
      //
      internal_change = false;
    }
  }
  
  void slot_button_toggled() {
    if (!internal_change) {
      internal_change = true;
      //
      *obj = isOn();
      //
      internal_change = false;
    }
  }
  
 private:
  BoolObject * const obj;
 private: 
  bool internal_change;
};


class XOrsaSizeSliderTool : public QSlider {
  
  Q_OBJECT
    
 public:
  XOrsaSizeSliderTool(SizeObject * size_obj, QWidget * parent) : QSlider(parent), obj(size_obj) { 
        
    internal_change = false;
    //
    setMinimumWidth(50);
    // 
    setMinValue(0);
    setMaxValue(obj->Size()-1);
    setValue(0);
    //
    slot_object_changed();
    //
    connect(obj,SIGNAL(changed()),this,SLOT(slot_object_changed()));
    connect(obj,SIGNAL(size_changed()),this,SLOT(slot_object_size_changed()));
    //
    connect(this,SIGNAL(valueChanged(int)),this,SLOT(slot_slider_value_changed(int)));
    connect(this,SIGNAL(sliderPressed()),this,SLOT(slot_slider_pressed()));
    connect(this,SIGNAL(sliderReleased()),this,SLOT(slot_slider_released()));
  }
  
 private slots:
  void slot_object_changed() {
    if (!internal_change) {
      internal_change = true;
      //
      setValue(*obj);
      //
      internal_change = false;
    }
  }
  
  void slot_object_size_changed() {
    if (!internal_change) {
      internal_change = true;
      //
      setMaxValue(obj->Size()-1);
      //
      internal_change = false;
    }
  }
  
  void slot_slider_value_changed(int i) {
    if (!internal_change) {
      internal_change = true;
      //
      *obj = i;
      //
      internal_change = false;
    }
  }
  
 protected slots:
  virtual void slot_slider_pressed() {
    focusWidget()->clearFocus();
    setFocus();
  }
  
  virtual void slot_slider_released() {
    
  }
  
 private:
  SizeObject * const obj;
 private: 
  bool internal_change;
};



class XOrsaAnimationDelaySliderTool : public QSlider {
  
  Q_OBJECT
    
 public:
  XOrsaAnimationDelaySliderTool(IntObject * int_obj, QWidget * parent) : QSlider(parent), obj(int_obj) { 
        
    internal_change = false;
    //
    setMinimumWidth(50);
    // 
    // the slider values are in FPS
    setMinValue(1);
    setMaxValue(120);
    //
    setValue(25);
    //
    slot_object_changed();
    //
    connect(obj,SIGNAL(changed()),this,SLOT(slot_object_changed()));
    //
    connect(this,SIGNAL(valueChanged(int)),this,SLOT(slot_slider_value_changed(int)));
    connect(this,SIGNAL(sliderPressed()),this,SLOT(slot_slider_pressed()));
    connect(this,SIGNAL(sliderReleased()),this,SLOT(slot_slider_released()));
  }
  
 private:
  int slider_from_ms(const int ms) {
    return ((int)rint(1000.0/ms));
  }
  
  int ms_from_slider(const int slider) {
    return ((int)rint(1000.0/slider));
  }
  
 private slots:
  void slot_object_changed() {
    if (!internal_change) {
      internal_change = true;
      //
      setValue(slider_from_ms(*obj));
      //
      internal_change = false;
    }
  }
  
  void slot_slider_value_changed(int i) {
    if (!internal_change) {
      internal_change = true;
      //
      *obj = ms_from_slider(i);
      //
      internal_change = false;
    }
  }
  
 protected slots:
  virtual void slot_slider_pressed() {
    focusWidget()->clearFocus();
    setFocus();
  }
  
  virtual void slot_slider_released() {
    
  }
  
 private:
  IntObject * const obj;
 private: 
  bool internal_change;
};


class XOrsaDoubleObjectWithLimitsSliderTool : public QSlider {
  
  Q_OBJECT
    
 public:
  XOrsaDoubleObjectWithLimitsSliderTool(DoubleObjectWithLimits * double_obj, QWidget * parent) : QSlider(parent), obj(double_obj) { 
        
    internal_change = false;
    //
    setMinimumWidth(50);
    // 
    setMinValue(0);
    setMaxValue(1000);
    //
    slot_object_changed();
    //
    connect(obj,SIGNAL(changed()),this,SLOT(slot_object_changed()));
    connect(obj,SIGNAL(limits_changed()),this,SLOT(slot_object_changed()));
    //
    connect(this,SIGNAL(valueChanged(int)),this,SLOT(slot_slider_value_changed(int)));
    connect(this,SIGNAL(sliderPressed()),this,SLOT(slot_slider_pressed()));
    connect(this,SIGNAL(sliderReleased()),this,SLOT(slot_slider_released()));
  }
  
 private:
  double factor() const {
    /* 
       std::cerr << "factor: " << pow((obj->max()/obj->min()),1.0/(maxValue()-minValue())) << std::endl;
       std::cerr << "obj->max(): " << obj->max() << std::endl;
       std::cerr << "obj->min(): " << obj->min() << std::endl;
       std::cerr << "maxValue(): " << maxValue() << std::endl;
       std::cerr << "minValue(): " << minValue() << std::endl;
    */
    //
    if (obj->min()==0.0) {
      ORSA_ERROR_MACRO("WARNING: obj->min() == 0.0");
    }
    return pow((obj->max()/obj->min()),1.0/(maxValue()-minValue()));
  }
  
  double double_from_int(const int i) const {
    return (obj->min()*pow(factor(),i-minValue()));
  }
  
  int int_from_double(const double d) const {
    return ((int)floor(log(d/obj->min())/log(factor()))+minValue());
  }
  
 private slots:
  void slot_object_changed() {
    if (!internal_change) {
      internal_change = true;
      //
      setValue(int_from_double(*obj));
      //
      internal_change = false;
    }
  }
  
  void slot_slider_value_changed(int i) {
    if (!internal_change) {
      internal_change = true;
      //
      *obj = double_from_int(i);
      //
      internal_change = false;
    }
  }
  
 protected slots:
  virtual void slot_slider_pressed() {
    focusWidget()->clearFocus();
    setFocus();
  }
  
  virtual void slot_slider_released() {
    
  }
  
 private:
  DoubleObjectWithLimits * const obj;
 private: 
  bool internal_change;
};



class XOrsaImprovedObjectsComboTool : public XOrsaImprovedObjectsCombo {
  
  Q_OBJECT
  
 public:
  XOrsaImprovedObjectsComboTool(IntObject * int_obj, QWidget * parent=0) : XOrsaImprovedObjectsCombo(parent), obj(int_obj) {
    init();
  }
  
  XOrsaImprovedObjectsComboTool(IntObject * int_obj, int special_objects_flags=0, QWidget * parent=0) : XOrsaImprovedObjectsCombo(special_objects_flags,parent), obj(int_obj) {
    init();
  }
					
  XOrsaImprovedObjectsComboTool(IntObject * int_obj, const std::vector<orsa::Body> * bodies, const bool only_massive=false, QWidget * parent=0) : XOrsaImprovedObjectsCombo(bodies,only_massive,parent), obj(int_obj) {
    init();
  }
  
  XOrsaImprovedObjectsComboTool(IntObject * int_obj, const std::vector<orsa::Body> * bodies, const bool only_massive=false, int special_objects_flags=0, QWidget * parent=0) : XOrsaImprovedObjectsCombo(bodies,only_massive,special_objects_flags,parent), obj(int_obj) {
    init();
  }
  
 private:
  void init() {
    
    internal_change = false;
    
    slot_object_changed();
    
    connect(obj,SIGNAL(changed()),this,SLOT(slot_object_changed()));
    //
    connect(this,SIGNAL(ObjectChanged(int)),this,SLOT(slot_combo_changed(int))); // here the ObjectChanged signal refers to the XOrsaImprovedObjectsCombo...
  }
  
 private slots:
  void slot_object_changed() {
    if (!internal_change) {
      internal_change = true;
      //
      SetObject(*obj);
      //
      internal_change = false;
    }
  }
  
  void slot_combo_changed(int i) {
    if (!internal_change) {
      internal_change = true;
      //
      *obj = i;
      //
      internal_change = false;
    }
  }
  
 private:
  IntObject * const obj;
 private: 
  bool internal_change;
};


// a rewrite of XOrsaOpenGL
class XOrsaOpenGLWidget : public QGLWidget {
  
  Q_OBJECT
    
 public:
  XOrsaOpenGLWidget(QWidget *parent=0, WFlags f=0);
  XOrsaOpenGLWidget(int w, int h, QWidget *parent=0, WFlags f=0);
  
 private:
  void init();
  void init_signals_slots();
  
 protected:
  void initializeGL();
  void resizeGL(int, int);
  void paintGL();
  
 private slots:
  void slot_call_resizeGL() { 
    resizeGL(width(),height());
    updateGL();
  }
  
 protected:
  virtual void draw() = 0;
  
  // utils
 public:
  void glTranslate(const orsa::Vector &v) const { glTranslated(v.x,v.y,v.z); }
  /* 
     orsa::Vector glTranslate(const orsa::Vector &v) const {  
     
     double m_pre[16];
     glGetDoublev(GL_MODELVIEW_MATRIX,m_pre);
     //
     double inverted_pre[16];
     invert(m_pre,inverted_pre);
     //
     glTranslated(v.x,v.y,v.z);
     //
     double m_post[16];
     glGetDoublev(GL_MODELVIEW_MATRIX,m_post);
     //
     double w[3];
     //
     w[0] = m_post[12]-m_pre[12];
     w[1] = m_post[13]-m_pre[13];
     w[2] = m_post[14]-m_pre[14];
     //
     orsa::Vector q;
     //
     q.x = w[0]*inverted_pre[0]+w[1]*inverted_pre[4]+w[2]*inverted_pre[8];
     q.y = w[0]*inverted_pre[1]+w[1]*inverted_pre[5]+w[2]*inverted_pre[9];
     q.z = w[0]*inverted_pre[2]+w[1]*inverted_pre[6]+w[2]*inverted_pre[10];
     //
     // note: v-q is the difference between the translation requested and 
     //       the real OpenGL translation: the two differ for many reasons,
     //       especially because OpenGL internally uses floats... :(
     //
     
     // return (v-q);
     return q;
     }
  */
  
  void glVertex(const orsa::Vector &v) const { glVertex3d(v.x,v.y,v.z); }
  
  // text, with gl2ps support
  void renderText(int x, int y, const QString &str, const QFont &fnt) {
    QGLWidget::renderText(x,y,"",fnt);
    const int fontsize=MAX(fnt.pixelSize(),fnt.pointSize());
    gl2psText(str,"Times-Roman",fontsize);
    QGLWidget::renderText(x,y,str,fnt);
  }
  
  void renderText(double x, double y, double z, const QString &str, const QFont &fnt) {
    QGLWidget::renderText(x,y,z,"",fnt);
    const int fontsize=MAX(fnt.pixelSize(),fnt.pointSize());
    gl2psText(str,"Times-Roman",fontsize);
    QGLWidget::renderText(x,y,z,str,fnt);  
  }
  
  void renderText(const orsa::Vector &v, const QString &str, const QFont &fnt) {
    renderText(v.x,v.y,v.z,str,fnt);
  }
  
  void renderText(int x, int y, const QString &str) {
    renderText(x,y,str,font());
  }
  
  void renderText(const orsa::Vector &v, const QString &str) {
    renderText(v,str,font());
  }
  
  //! length units when projection==OGL_ORTHO,
  //! degrees when projection==OGL_PERSPECTIVE.
  double PixelLength() const {
    double pixel_length = 0.0;
    switch(projection) {
    case OGL_ORTHO:
      pixel_length = 2.0*ortho_xy_scale/MIN(height(),width());
      break;
    case OGL_PERSPECTIVE:
      pixel_length = FOV/height(); // FOV is always along y (see man gluPerspective)
      break;
    }
    return pixel_length;
  }
  
 public:
  void export_file(const QString &, const OpenGL_Export_File);
 public slots:
  // handy export slots...
  void export_png();
  void export_ps();
  void export_pdf();
  
  // camera
 public:
  ProjectionObject projection;
  //
  DoubleObjectWithLimits distance;
  //
  DoubleObjectWithLimits FOV;
  DoubleObjectWithLimits near;
  DoubleObjectWithLimits far;
  //
  BoolObject near_and_far_limit_on_distance;
  //
  DoubleObjectWithLimits ortho_xy_scale;
  DoubleObjectWithLimits ortho_z_near_scale;
  DoubleObjectWithLimits ortho_z_far_scale;
  //
 private slots:
  void slot_near_changed();
  void slot_far_changed();
 private:
  bool bool_near_internal_change;
  bool bool_far_internal_change;
 private slots:
  void slot_near_limits_changed();
  void slot_far_limits_changed();
 private:
  bool bool_near_limits_internal_change;
  bool bool_far_limits_internal_change;
 private slots:
  void slot_near_and_far_limit_on_distance_changed();  
  
  // animation
 protected slots:
  virtual void animate();
 private slots:
  void slot_bool_animate_changed();
  void slot_animation_delay_changed();
 public:
  BoolObject bool_animate;
  IntObject  animation_delay_ms;
 protected:
  bool bool_already_animating;
 private:
  QTimer animation_timer;
 protected:
  void showEvent(QShowEvent*);
  void hideEvent(QHideEvent*);
  
  // center rotation
 public:
  DoubleObjectPeriodic center_X_rotation;
  DoubleObjectPeriodic center_Y_rotation;
  DoubleObjectPeriodic center_Z_rotation;
  //
  DoubleObjectWithLimits center_rotation_impulse;
  
  // eye rotation
 public:
  DoubleObjectPeriodic eye_X_rotation;
  DoubleObjectPeriodic eye_Y_rotation;
  DoubleObjectPeriodic eye_Z_rotation;
  //
  DoubleObjectWithLimits eye_rotation_impulse;
 private slots:
  void slot_update_eye_rotation_impulse();
  
 protected:
  void mousePressEvent(QMouseEvent*);
  void mouseReleaseEvent(QMouseEvent*);
  void mouseMoveEvent(QMouseEvent*);
  void wheelEvent(QWheelEvent*);
 private:
  QPoint old_mouse_position;
  
  // FPS
 private:
  FPS fps;
  int last_fps_frame_index;
 public:
  // get_fps: used by OSD, allows to keep FPS private...
  double get_fps() const {
    return fps; 
  }
};


class OSD;

class XOrsaOpenGLEvolutionWidget : public XOrsaOpenGLWidget {
  
  Q_OBJECT
  
 public:
  XOrsaOpenGLEvolutionWidget(QWidget *parent=0, WFlags f=0);
  XOrsaOpenGLEvolutionWidget(int w, int h, QWidget *parent=0, WFlags f=0);
  
 private:
  void init();
  
  // Evolution
 public:
  void SetEvolution(const orsa::Evolution*);
 private:
  const orsa::Evolution * evolution; 
  orsa::Frame evol_frame;
 public:
  SizeObject evol_counter;
 private:
  void update_sizes();
 signals:
  void evolution_changed();
  void frame_changed(int); 
  public slots:
 public:	
  unsigned int total_frames() const {
    if (evolution != 0) return evolution->size();
    return 0;
  }
 private:
  bool frame_changed_from_animation;
  
 private:
  void customEvent(QCustomEvent*);
  
 public:
  const orsa::Frame * bodies() const { 
    // should check more...
    if (evolution)
      return &(evolution[0][0]);
    else 
      return 0;
  };
  
 public slots:
  void export_movie();
  
  // scale
 private:
  DoubleObject range;
  unsigned int last_size_checked;
  void update_range();
 private slots:
  void range_changed() {
    distance.SetMax( 10.0*range);
    near.SetMax(     11.0*range);
    far.SetMax(      11.0*range);
    //
    ortho_xy_scale.SetMax(     11.0*range);
    ortho_z_near_scale.SetMax( 11.0*range);
    ortho_z_far_scale.SetMax(  11.0*range);
  }
  
 protected:
  void initializeGL();
  
 private:
  void draw();
  
 private:
  GLint earth, earth_vector;
  GLint moon;
  GLint circle;
  GLint reference_system;
  GLint xy_grid;
  //
  int max_circle_index;
  
  /* 
     private:
     GLuint object_names_base_list;
     void generate_object_names();
  */
  
 protected:
  void animate();
  
  // reference frame: center, eye and rotation
 public:
  IntObject  center_body;
  //
  IntObject  eye_body;
  BoolObject eye_on_body; // the eye position is locked on the eye_body;
  //
  IntObject  rotation_body;
  BoolObject rotate_with_rotation_body;
  
 private slots:
  void eye_body_changed() {
    if (eye_on_body) {
      rotation_body = eye_body;
      rotate_with_rotation_body = true;
      
      if ((eye_body >= 0) && (eye_body < (int)evol_frame.size())) {
	const double eye_body_radius = evol_frame[eye_body].radius();
	if (eye_body_radius > 0.0) {
	  // near.SetMin(1.05*eye_body_radius);
       	  near.SetMin(0.05*eye_body_radius);
       	  //
	  ortho_z_near_scale.SetMin(1.05*eye_body_radius);
	}
      }      
    }
  }
  
  void eye_on_body_changed() {
    // TEST: disconnect updateGL() from distance.changed() when
    // using eye_on_body, as the distance in this case is not
    // free. This should solve the problem of having paintGL()
    // called multiple times when eye_on_body is true.
    
    if (eye_on_body) {
      disconnect(&distance,SIGNAL(changed()),this,SLOT(updateGL()));
    } else {
      connect(&distance,SIGNAL(changed()),this,SLOT(updateGL()));
    } 
    
    // angles update
    if (eye_on_body) {
      center_X_rotation.Lock(270.0);
      center_Y_rotation.Lock(000.0);
      center_Z_rotation.Lock(090.0);
      //
      rotation_body = eye_body;
      rotate_with_rotation_body = true;
    } else {
      center_X_rotation.UnLock();
      center_Y_rotation.UnLock();
      center_Z_rotation.UnLock();
      //
      // rotate_with_rotation_body = false;
    }
  }
  
 private:
  orsa::Vector BodyPosition(const int) const;
  //
  orsa::Vector CenterBodyPosition() const;
  orsa::Vector EyeBodyPosition() const;
  orsa::Vector RotationBodyPosition() const;
  //
  orsa::Vector EyePosition() const { return eye_position; }
  //
  void  update_EyePosition();
  orsa::Vector eye_position;
  //
  std::string  BodyName(const int) const;
  //
  std::string  CenterBodyName() const;
  std::string  EyeBodyName() const;
  std::string  RotationBodyName() const;
  
 private slots:
  void evol_counter_changed() {
    // std::cerr << "        --------> evol_counter = " << evol_counter << std::endl;
    evol_frame = (*evolution)[evol_counter];
  }
  
  void center_body_changed() {
    
    eye_X_rotation = eye_Y_rotation = eye_Z_rotation = 0.0;
    
    if (center_body >= 0) {
      const double center_body_radius = evol_frame[center_body].radius();
      if (center_body_radius > 0.0) {
	ortho_xy_scale.SetMin(     1.05*center_body_radius);
	//
	ortho_z_near_scale.SetMin( 1.05*center_body_radius);
	ortho_z_far_scale.SetMin(  1.05*center_body_radius); }
    }
  }
  
  void update_distance() {
    
    if (projection != OGL_PERSPECTIVE) return;
    
    if (center_body >= 0) {
      const double center_body_radius = evol_frame[center_body].radius();
      if (center_body_radius > 0.0) {
	distance.SetMin(1.10*center_body_radius);
      }
    }
    
    if ((eye_on_body) && (center_body != eye_body)) {
      distance = (EyeBodyPosition() - CenterBodyPosition()).Length();
    }    
  }
  
  void distance_changed() {
    
    if (near_and_far_limit_on_distance) {
      if ((center_body >= 0) && (center_body < (int)evol_frame.size())) {
	const double center_body_radius = evol_frame[center_body].radius();
	if (center_body_radius > 0.0) {
	  // this min should be already set, but...
	  distance.SetMin(1.10*center_body_radius);
	  //
	  near.SetMax(distance-1.05*center_body_radius);
	  far.SetMin( distance+1.05*center_body_radius);
	} else {
       	  near.SetMax(distance);
	  far.SetMin(distance);
	}
      } else {
	near.SetMax(distance);
	far.SetMin(distance);
      }
    }
  }
  
  // orbits
 public:
  BoolObject draw_orbits;
  IntObject orbit_reference_body_index;
  BoolObject bright_positive_Z;
 private:
  mutable std::vector < std::vector <int> > orbit_reference_body_index_vector;
  mutable std::vector < std::map < int , OrbitCache > > orbit_cache_vector;
  
  // MOID
 public:
  BoolObject draw_MOID;
  IntObject MOID_body_1;
  IntObject MOID_body_2;
  
  // Lagrange points
 public:
  BoolObject draw_Lagrange_points;
  IntObject Lagrange_body_1;
  IntObject Lagrange_body_2;
  
  // tracks
 public:
  BoolObject draw_tracks;
  BoolObject absolute_tracks;
  
  // misc
 public:
  BoolObject draw_reference_system;
  BoolObject draw_XY_grid;
  BoolObject draw_ra_dec_grid;
  BoolObject draw_Z_bars;
  BoolObject draw_scale_bar;
  BoolObject light_time_correction;
  
  // labels
 public:
  BoolObject       draw_labels;
  LabelsModeObject labels_mode;

 private:
  void internal_draw_bodies_and_labels();
  void internal_draw_orbit_ellipse();
  void internal_draw_MOID();
  void internal_draw_Lagrange_points();
  //
  void internal_draw_OSD(); // On Screen Display
  friend class OSD;
  
};

enum OSD_ZONE {
  TOP_LEFT,
  TOP_CENTER,
  TOP_RIGHT,
  CENTER_LEFT,
  CENTER,
  CENTER_RIGHT,
  BOTTOM_LEFT,
  BOTTOM_CENTER,
  BOTTOM_RIGHT
};

class OSDZoneObject : public QObject {
  
  Q_OBJECT
  
 public:
  OSDZoneObject() : QObject() { 
    _z = TOP_LEFT;
  }
  
  OSDZoneObject(const OSD_ZONE z) : QObject() { 
    _z = z;
  }
  
 public:
  OSD_ZONE operator = (const OSD_ZONE z) {
    if (z != _z) {
      _z = z;
      emit changed();
    }
    return _z;
  }
  
  operator OSD_ZONE() const {
    return _z;
  }	
  
 signals:
  void changed();
  
 private:
  OSD_ZONE _z;  
};

class OSD : public QObject {
  
  Q_OBJECT
  
 public:
  OSD(XOrsaOpenGLEvolutionWidget * opengl_ew) : opengl(opengl_ew), fm(opengl->font()) {
    connect(&zone, SIGNAL(changed()), this, SLOT(zone_changed()));
    zone = TOP_LEFT;
    x = y = 0;
  }
  
 public:
  void draw();
  
 private slots:
  void zone_changed() {
    const int w  = opengl->width();
    const int h  = opengl->width();
    const int w2 = opengl->width()/2;
    const int h2 = opengl->width()/2;
    switch (zone) {
    case TOP_LEFT:      x = 0;  y = 0;  break;
    case TOP_CENTER:    x = w2; y = 0;  break;
    case TOP_RIGHT:     x = w;  y = 0;  break;
    case CENTER_LEFT:   x = 0;  y = h2; break;
    case CENTER:        x = w2; y = h2; break;
    case CENTER_RIGHT:  x = w;  y = h2; break;
    case BOTTOM_LEFT:   x = 0;  y = h;  break;
    case BOTTOM_CENTER: x = w2; y = h;  break;
    case BOTTOM_RIGHT:  x = w;  y = h;  break;
    }
  }
  
 private:
  void write(const QString & str) {
    const QRect str_rect = fm.boundingRect(str);
    //
    x = MAX(x,abs(str_rect.left())+fm.maxWidth()); // ..if close to left border
    x = MIN(x,opengl->width()-str_rect.right()-fm.maxWidth()); // ..if close to right border
    //
    y = MAX(y,fm.lineSpacing()); // ..if close to top border
    y = MIN(y,opengl->height()-fm.lineSpacing()); // ..if close to right border
    //
    opengl->renderText(x,y,str);
    //
    switch (zone) {
    case TOP_LEFT:
    case TOP_CENTER: 
    case TOP_RIGHT:
      y += fm.lineSpacing();
      break;
    case CENTER_LEFT:
    case CENTER:
    case CENTER_RIGHT:
      y += fm.lineSpacing();
      break;
    case BOTTOM_LEFT:
    case BOTTOM_CENTER:
    case BOTTOM_RIGHT:
      y -= 2 * fm.lineSpacing();
      break;
    }
  }
  
 public:
  OSDZoneObject zone;
  
 private:
  // cursor position...
  int x, y;
  
 private:
  XOrsaOpenGLEvolutionWidget * const opengl;
  
 private:
  const QFontMetrics fm;
};









class XOrsaOpenGL : public GLControlWidget {
  
  Q_OBJECT
  
 public:
  XOrsaOpenGL(QWidget *parent=0, WFlags f=0);
  XOrsaOpenGL(int w, int h, QWidget *parent=0, WFlags f=0);
  
  
 private:
  void init();
  
  
  // Evolution
 public:
  void SetEvolution(const orsa::Evolution*);
 public:
  //  private:
  // could be private, fix it!
  const orsa::Evolution *evolution; 
 signals:
  void evolution_size_changed(int);
  void evolution_changed();
  
  
  // Animation 
 public:
  inline void StartAnimation()    { bool_animate=true;   };
  inline void StopAnimation()     { bool_animate=false;  };
  inline bool IsAnimating() const { return bool_animate; };
 private:
  bool bool_animate;
  bool bool_already_animating;
  
  
  // Orbits
 public slots: 
  inline void DrawOrbits(bool b) { bool_draw_orbits = b; updateGL(); };
 public:
  inline bool DrawOrbits() const { return (bool_draw_orbits); };
 private:
  bool bool_draw_orbits;
 private:
  void internal_draw_orbit_ellipse();
 signals:
  void orbit_reference_body_changed(int); 
 public slots:
  inline void orbit_reference_body(int i) {
    orbit_reference_body_index = i;
    emit orbit_reference_body_changed(i);
    updateGL();
  }
 private:
  std::vector < std::vector <int> > orbit_reference_body_index_vector;
  std::vector < std::map <int,OrbitCache> > orbit_cache_vector;
 private:
  int orbit_reference_body_index;
  
  
  // Bright Positive Z
 public slots: 
  inline void BrightPositiveZ(bool b) { bool_bright_positive_z = b; updateGL(); };
 public:
  inline bool BrightPositiveZ() const { return (bool_bright_positive_z); };
 private:
  bool bool_bright_positive_z;
  
  
  // MOID
 public slots: 
  inline void DrawMOID(bool b) { bool_draw_moid = b; updateGL(); };
 public:
  inline bool DrawMOID() const { return (bool_draw_moid); };
 private:
  bool bool_draw_moid;
 public slots:
  inline void MoidBody1(int i) { moid_body_1 = i; updateGL(); };
  inline void MoidBody2(int i) { moid_body_2 = i; updateGL(); };
 private:
  void internal_draw_MOID();
 private:
  int moid_body_1, moid_body_2;
  
  
  // Lagrange points
 public slots: 
  inline void DrawLagrange(bool b) { bool_draw_lagrange = b; updateGL(); };
 public:
  inline bool DrawLagrange() const { return (bool_draw_lagrange); };
 private:
  bool bool_draw_lagrange;
 public slots:
  inline void LagrangeBody1(int i) { lagrange_body_1 = i; updateGL(); };
  inline void LagrangeBody2(int i) { lagrange_body_2 = i; updateGL(); };
 private:
  void internal_draw_Lagrange_points(orsa::Vector&);
 private:
  int lagrange_body_1, lagrange_body_2;
  
  
  // Reference system
 public slots: 
  inline void DrawReferenceSystem(bool b) { bool_draw_reference_system  = b; updateGL(); };
 public:
  inline bool DrawReferenceSystem() const { return (bool_draw_reference_system); };
 private:
  bool bool_draw_reference_system;
 private:
  void internal_draw_reference_system();
 private:
  GLint reference_system;
  

  // XY grid
 public slots: 
  inline void DrawXYGrid(bool b) { bool_draw_xy_grid = b; updateGL(); };
 public:
  inline bool DrawXYGrid() const { return (bool_draw_xy_grid); };
 private:
  bool bool_draw_xy_grid;
 private:
  void internal_draw_xy_grid();
 private:
  GLint xy_grid;
  
  
  // Z bars
 public slots: 
  inline void DrawZBars(bool b) { bool_draw_z_bars = b; updateGL(); };
 public:
  inline bool DrawZBars() const { return (bool_draw_z_bars); };
 private:
  bool bool_draw_z_bars;
 private:
  void internal_draw_z_bars();
  
  
  // Tracks
 public slots: 
  inline void DrawTracks(bool b) { bool_draw_tracks = b; updateGL(); };
 public:
  inline bool DrawTracks() const { return (bool_draw_tracks); };
 private:
  bool bool_draw_tracks;
 private:
  void internal_draw_track();
  
  
  // Absolute Tracks
 public slots: 
  inline void AbsoluteTracks(bool b) { bool_absolute_tracks = b; updateGL(); };
 public:
  inline bool AbsoluteTracks() const { return (bool_absolute_tracks); };
 private:
  bool bool_absolute_tracks;
  
  
  // Draw Bodies and Labels
 public slots: 
  inline void DrawLabels(bool b) { bool_draw_labels = b; updateGL(); };
 public:
  inline bool DrawLabels() const { return (bool_draw_labels); };
 private:
  bool bool_draw_labels;
 private:
  void internal_draw_bodies_and_labels();
  
  
  // Scale Bar
 public slots: 
  inline void DrawScaleBar(bool b) { bool_draw_scale_bar = b; updateGL(); };
 public:
  inline bool DrawScaleBar() const { return (bool_draw_scale_bar); };
 private:
  bool bool_draw_scale_bar; 
 private:
  void internal_draw_scalebar();
  
  
  // Draw FPS
 public slots: 
  inline void DrawFPS(bool b) { bool_draw_FPS = b; updateGL(); };
 public:
  inline bool DrawFPS() const { return (bool_draw_FPS); };
 private:
  bool bool_draw_FPS;
 private:
  void internal_draw_FPS();
 private:
  double fps;
  int last_fps_frame_index;
  QTime qtime;
  std::vector<double> fps_running_average_vector;
  unsigned int   fps_running_average_index;
  
  
  // Direction (various)
  public slots: 
  inline void UseDirectionBody(bool b) { bool_use_direction_body = b; updateGL(); };
  inline void ObserveFromDirectionBody(bool b) { bool_observe_from_direction_body = b; updateGL(); };
   public:
  inline bool UseDirectionBody() const { return (bool_use_direction_body); };
  inline bool ObservingFromDirectionBody() const { return (bool_observe_from_direction_body); };
 private:
  bool bool_use_direction_body;
  bool bool_observe_from_direction_body;
 public slots:
  inline void direction_body(int i) {
    direction_body_index = i;
    emit direction_changed(i);
    if (projection == OGL_PERSPECTIVE) minus_YZ();
    updateGL();
  };
 public:
  std::string direction_body_name() const;
 private:
  void internal_draw_direction_line(orsa::Vector&);
 signals:
  void direction_changed(int);
 private:
  int direction_body_index;
  
  
  // Light-Time Correction
 public slots: 
  inline void LightTimeCorrection(bool b) { bool_light_time_correction  = b; updateGL(); };
 public:
  inline bool LightTimeCorrection() const { return (bool_light_time_correction); };
 private:
  bool bool_light_time_correction;
  
  
  
 public slots: 
  // inline void DrawOrbits(bool b)          { bool_draw_orbits            = b; updateGL(); };
  // inline void DrawMOID(bool b)            { bool_draw_moid              = b; updateGL(); };
  // inline void DrawReferenceSystem(bool b) { bool_draw_reference_system  = b; updateGL(); };
  // inline void DrawXYGrid(bool b)          { bool_draw_xy_grid           = b; updateGL(); };
  // inline void DrawZBars(bool b)           { bool_draw_z_bars            = b; updateGL(); };
  // inline void DrawTracks(bool b)          { bool_draw_tracks            = b; updateGL(); };
  // inline void DrawLabels(bool b)          { bool_draw_labels            = b; updateGL(); };
  // inline void DrawLagrange(bool b)        { bool_draw_lagrange          = b; updateGL(); };
  // inline void DrawScaleBar(bool b)        { bool_draw_scale_bar         = b; updateGL(); };
  // inline void DrawFPS(bool b)             { bool_draw_FPS               = b; updateGL(); };
  // inline void UseDirectionBody(bool b)    { bool_use_direction_body     = b; updateGL(); };
  // inline void BrightPositiveZ(bool b)     { bool_bright_positive_z      = b; updateGL(); };
  // inline void LightTimeCorrection(bool b) { bool_light_time_correction  = b; updateGL(); };
  // inline void AbsoluteTracks(bool b)      { bool_absolute_tracks        = b; updateGL(); };
  //
  // inline void ObserveFromDirectionBody(bool b) { bool_observe_from_direction_body = b; updateGL(); };
  
 public:
  // inline bool DrawOrbits()          const { return (bool_draw_orbits);             };
  // inline bool DrawMOID()            const { return (bool_draw_moid);               };
  // inline bool DrawReferenceSystem() const { return (bool_draw_reference_system);   };
  // inline bool DrawXYGrid()          const { return (bool_draw_xy_grid);            };
  // inline bool DrawZBars()           const { return (bool_draw_z_bars);             };
  // inline bool DrawTracks()          const { return (bool_draw_tracks);             };
  // inline bool DrawLabels()          const { return (bool_draw_labels);             };
  // inline bool DrawLagrange()        const { return (bool_draw_lagrange);           };
  // inline bool DrawScaleBar()        const { return (bool_draw_scale_bar);          };
  // inline bool DrawFPS()             const { return (bool_draw_FPS);                };
  // inline bool UseDirectionBody()    const { return (bool_use_direction_body);      };
  // inline bool BrightPositiveZ()     const { return (bool_bright_positive_z);       };
   // inline bool LightTimeCorrection() const { return (bool_light_time_correction);   };
  // inline bool AbsoluteTracks()      const { return (bool_absolute_tracks);         };
  //
  // inline bool ObservingFromDirectionBody() const { return (bool_observe_from_direction_body); };
  
 private:
  // bool bool_use_direction_body;
  // bool bool_bright_positive_z;
  // bool bool_light_time_correction;
  // bool bool_draw_scale_bar; 
  // bool bool_draw_FPS;
  // bool bool_draw_lagrange;
  // bool bool_draw_orbits, bool_draw_moid, bool_draw_tracks, bool_draw_labels, bool_draw_reference_system;
  // bool bool_draw_tracks, bool_draw_labels, bool_draw_reference_system;
  // bool bool_draw_labels;
  // bool bool_draw_z_bars;
  // bool bool_absolute_tracks;
  //
  // bool bool_observe_from_direction_body;
  
  
 public slots: 
  inline void XY() { 
    xRot = yRot = zRot = 0;
    updateGL();
  };
  
  inline void XZ() { 
    yRot = zRot = 0;
    xRot = 270;
    updateGL();
  };
  
  inline void YZ() { 
    yRot = 0;
    xRot = zRot = 270;
    updateGL();
  };
  
  inline void minus_YZ() { 
    yRot = 0;
    xRot = 270;
    zRot = 90;
    updateGL();
  };
  
  
 public slots:
  inline void center_on_body(int i) {
    central_body_index = i;
    xTrans = yTrans = zTrans = 0.0;
    emit center_changed(i);
    updateGL();
  };
  
  /* 
     inline void orbit_center_body(int i) {
     orbit_center_body_index = i;
     emit orbit_center_changed(i);
     updateGL();
     };
  */
  
  /* 
     inline void MoidBody1(int i) {
     moid_body_1 = i;
     updateGL();
     };
     
     inline void MoidBody2(int i) {
     moid_body_2 = i;
     updateGL();
     };
  */
  
  /* 
     inline void LagrangeBody1(int i) {
     lagrange_body_1 = i;
     updateGL();
     };
     
     inline void LagrangeBody2(int i) {
     lagrange_body_2 = i;
     updateGL();
     };
  */
  
  /* 
     inline void direction_body(int i) {
     direction_body_index = i;
     emit direction_changed(i);
     // if (projection == OGL_PERSPECTIVE) YZ();
     if (projection == OGL_PERSPECTIVE) minus_YZ();
     updateGL();
     };
  */
  
  inline void setZoomScaleZero(double zsz) {
    zoom_scale_zero = zsz;
    updateFOV();
    updateGL();
  }
  
  inline void updateFOV() {
    switch (projection) {
    case OGL_ORTHO:       fov=zoom_scale_zero*fov_base; break;
    case OGL_PERSPECTIVE: fov=150.0*fov_base;           break;
    }
  }
  
  inline void SetLabelsMode(OpenGL_Labels_Mode lm) {
    labels_mode = lm;
    updateGL();
  }
  
  inline void SetProjection(OpenGL_Projection p) {
    projection = p;
    // if (projection == OGL_PERSPECTIVE) YZ();
    resizeGL(width(),height());
    updateGL();
    emit projection_changed();
  }
  
  inline OpenGL_Projection GetProjection() const {
    return projection;
  }
  
 public:
  std::string center_body_name() const;
  // std::string direction_body_name() const;
  
 protected:
  void initializeGL();
  void resizeGL(int, int);
  void paintGL();
  void draw();
  void animate();
  
 public:
  void export_file(const QString &s_filename, const OpenGL_Export_File);
  
 public:
  inline const orsa::Frame * bodies() const { return &evolution[0][0]; };
  
 private slots:
  void mouse_slot(int,int);
  
 private:
  // void internal_draw_direction_line(Vector&);
  // void internal_draw_Lagrange_points(Vector&);
  // void internal_draw_reference_system();
  // void internal_draw_xy_grid();
  // void internal_draw_z_bars();
  // void internal_draw_MOID();
  // void internal_draw_bodies_and_labels();
  // void internal_draw_track();
  // void internal_draw_orbit_ellipse();
  void internal_draw_time_label();
  // void internal_draw_scalebar();
  // void internal_draw_FPS();
  void internal_draw_distance();
  
 signals:
  void frame_changed(int);
  // void evolution_size_changed(int);
  // void evolution_changed();
  void center_changed(int);
  // void direction_changed(int);
  // void orbit_reference_body_changed(int);
  void projection_changed();
  
 public slots:
  void set_frame(int);
  void setFOV(int);
  void setDistance(int); // useful when projection == OGL_PERSPECTIVE
  
 public:	
  unsigned int total_frames() const;
  
 private:
  bool frame_changed_from_animation;
  
 private:
  // GLint planet,circle,reference_system;
  // GLint planet, reference_system, xy_grid, earth;
  GLint planet, earth;
  
  int max_circle_index;
  GLint circle;
  
  // text, with gl2ps support
 private:
  void renderText(int x, int y, const QString &str, const QFont &fnt); 
  void renderText(double x, double y, double z, const QString &str, const QFont &fnt); 
  
 private:
  // inverted modelview matrix
  double inverted[16];
  int mouse_x, mouse_y;
  
 private:
  OpenGL_Labels_Mode labels_mode;
  OpenGL_Projection  projection;
  
 private:
  // Evolution::iterator evol_it;
  orsa::Frame evol_frame;
  int evol_counter;
  
  
  /* 
     public slots:
     inline void orbit_reference_body(int i) {
     orbit_reference_body_index = i;
     emit orbit_reference_body_changed(i);
     updateGL();
     }
     
     private:
     vector < vector <int> > orbit_reference_body_index_vector;
     vector < std::map <int,OrbitCache> > orbit_cache_vector;
  */
  
 private:
  // int central_body_index, orbit_reference_body_index, direction_body_index;
  int central_body_index;
  // int moid_body_1, moid_body_2;
  
  
 private:
  double fov;
  double fov_base;
  double zoom_scale_zero;
  //
  double distance;
  //
  /* 
     double fps;
     int last_fps_frame_index;
     QTime qtime;
     vector<double> fps_running_average_vector;
     unsigned int   fps_running_average_index;
  */
  
 private:
  void UpdateVectors();
  orsa::Vector dir_pos,dir_pos_mod,center,rp,dir_pos_zero;
  
 private:
  QFont default_font;
  
 public slots:
  void UpdateLocation(const orsa::Location&);
  
 private:
  orsa::Location location;
};




class XOrsaOpenGLEvolutionTool : public QMainWindow {
  
  Q_OBJECT
    
 public:
  XOrsaOpenGLEvolutionTool(QWidget *parent=0);
  XOrsaOpenGLEvolutionTool(int w, int h, QWidget *parent=0);
  
  ~XOrsaOpenGLEvolutionTool();
  
 public:
  void SetEvolution(const orsa::Evolution*);
  
 private:
  void init_toolbars();
  
 private slots:
  void slot_set_orbit_reference_body_index(int);
  void slot_set_moid_body_1_index(int);
  void slot_set_moid_body_2_index(int);
  //
  void widgets_enabler();
  
 private:
  XOrsaBoolToolButton *plot_orbits_tb;
  // XOrsaAutoObjectsCombo *oc_orbits;
  XOrsaImprovedObjectsCombo *oc_orbits;
  XOrsaBoolToolButton *bright_positive_z_tb;
  //
  XOrsaBoolToolButton *moid_tb;
  // XOrsaObjectsCombo *oc_moid_1, *oc_moid_2;
  XOrsaImprovedObjectsCombo *oc_moid_1, *oc_moid_2;
  
 private:
  XOrsaOpenGLEvolutionWidget * opengl;
};







class TimeSlider : public XOrsaSizeSliderTool {
  
  Q_OBJECT
  
 public:
  TimeSlider(SizeObject * size_object, XOrsaOpenGLEvolutionWidget * opengl_ew, QWidget * parent=0) : XOrsaSizeSliderTool(size_object,parent), opengl(opengl_ew) {
    /* setWFlags(Qt::WRepaintNoErase);
       setBackgroundMode(Qt::NoBackground);
    */
  }
  
 protected slots:
  void slot_slider_pressed() {
    XOrsaSizeSliderTool::slot_slider_pressed();
    old_animation_state = opengl->bool_animate;
    opengl->bool_animate = false;
  }
  
  void slot_slider_released() {
    XOrsaSizeSliderTool::slot_slider_released();
    opengl->bool_animate = old_animation_state;
  }
  
 private:
  SizeObject * size_object;
  XOrsaOpenGLEvolutionWidget * const opengl;
 private:
  bool old_animation_state;
};











class XOrsaAnimationToolBar : public QToolBar {
 public:
  XOrsaAnimationToolBar(XOrsaOpenGLEvolutionWidget * opengl, QMainWindow * parent) : QToolBar(parent) {
    
    setLabel("Animation Tools");		  
    
    new XOrsaBoolToolButton(&opengl->bool_animate,QIconSet(play_xpm),"Start/Stop Animation",this); 
    
    TimeSlider * time_slider = new TimeSlider(&opengl->evol_counter,opengl,this);
    time_slider->setOrientation(Qt::Horizontal);
    QToolTip::add(time_slider,"Time Control");
    
    XOrsaAnimationDelaySliderTool * delay_slider = new XOrsaAnimationDelaySliderTool(&opengl->animation_delay_ms,this);
    delay_slider->setOrientation(Qt::Horizontal);
    QToolTip::add(delay_slider,"Animation Speed");
  }
};



class XOrsaLagrangePointsToolBar :  QToolBar {
  
  Q_OBJECT
    
 public:
  XOrsaLagrangePointsToolBar(XOrsaOpenGLEvolutionWidget * opengl_ew, QMainWindow * parent) : QToolBar(parent), opengl(opengl_ew) {
				  
    setLabel("Lagrange Points");		  
    
    enable_tb = new XOrsaBoolToolButton(&opengl->draw_Lagrange_points,QIconSet(lagrange_xpm),"Compute Lagrange Points Position",this); 
    connect(enable_tb,SIGNAL(toggled(bool)),this,SLOT(widgets_enabler()));
    
    combo_1 = new XOrsaImprovedObjectsComboTool(&opengl->Lagrange_body_1,0,this);
    QToolTip::add(combo_1,"Body 1");
    
    combo_2 = new XOrsaImprovedObjectsComboTool(&opengl->Lagrange_body_2,0,this);
    QToolTip::add(combo_2,"Body 2");
    
    connect(opengl,SIGNAL(evolution_changed()),this,SLOT(slot_evolution_changed()));
    
    widgets_enabler();
  } 
  
 private slots:
  void slot_evolution_changed() {
    const orsa::Frame * bodies = opengl->bodies();
    combo_1->Set(bodies,true);
    combo_2->Set(bodies,true);
  }
  
 private slots:  
  void widgets_enabler() {
    if (enable_tb->isOn()) {
      combo_1->setEnabled(true);
      combo_2->setEnabled(true);
    } else {
      combo_1->setEnabled(false);
      combo_2->setEnabled(false);
    }
  }
  
 private:
  XOrsaOpenGLEvolutionWidget * const opengl; 
  
 private:
  XOrsaBoolToolButton * enable_tb;
  
 private:
  XOrsaImprovedObjectsComboTool * combo_1;
  XOrsaImprovedObjectsComboTool * combo_2;
};

class XOrsaCameraToolBar : public QToolBar {
  
  Q_OBJECT
  
 public:
  XOrsaCameraToolBar(XOrsaOpenGLEvolutionWidget * opengl_ew, QMainWindow * parent) : QToolBar(parent), opengl(opengl_ew) {
			       
    setLabel("Camera Tools");
    
    center_cb = new XOrsaImprovedObjectsComboTool(&opengl->center_body, HEX_ORIGIN | HEX_CENTER_OF_MASS, this);
    QToolTip::add(center_cb,"Camera Center");
    
    connect(opengl,SIGNAL(evolution_changed()),this,SLOT(slot_evolution_changed()));
    
    distance_slider = new XOrsaDoubleObjectWithLimitsSliderTool(&opengl->distance,this);
    distance_slider->setOrientation(Qt::Horizontal);
    QToolTip::add(distance_slider,"Distance From Center");
    
    addSeparator();
    
    eye_tb = new XOrsaBoolToolButton(&opengl->eye_on_body,QIconSet(camera_position_xpm),"Enable/Disable Camera Position",this); 
    connect(eye_tb,SIGNAL(toggled(bool)),this,SLOT(widgets_enabler()));
    
    eye_cb = new XOrsaImprovedObjectsComboTool(&opengl->eye_body, HEX_ORIGIN | HEX_CENTER_OF_MASS,this);
    QToolTip::add(eye_cb,"Camera Position");
    
    addSeparator();
    
    rotation_tb = new XOrsaBoolToolButton(&opengl->rotate_with_rotation_body,QIconSet(rotation_xpm),"Enable/Disable Rotation",this); 
    connect(rotation_tb,SIGNAL(toggled(bool)),this,SLOT(widgets_enabler()));
    
    rotation_cb = new XOrsaImprovedObjectsComboTool(&opengl->rotation_body, HEX_ORIGIN | HEX_CENTER_OF_MASS, this);
    QToolTip::add(rotation_cb,"Rotation Body");
    
    addSeparator();
    
    projection_cb = new QComboBox(false,this);
    //
    projection_cb->insertItem("Orthographic");
    projection_cb->insertItem("Perspective");
    //
    connect(projection_cb,SIGNAL(activated(const QString &)),this,SLOT(slot_projection_changed()));
    //
    QToolTip::add(projection_cb,"Camera Projection");
    
    FOV_slider = new XOrsaDoubleObjectWithLimitsSliderTool(&opengl->FOV,this);
    FOV_slider->setOrientation(Qt::Horizontal);
    QToolTip::add(FOV_slider,"Field Of View");
    
    // or...
    
    ortho_xy_slider = new XOrsaDoubleObjectWithLimitsSliderTool(&opengl->ortho_xy_scale,this);
    ortho_xy_slider->setOrientation(Qt::Horizontal);
    QToolTip::add(ortho_xy_slider,"Scale");
    
    addSeparator();
    
    near_slider = new XOrsaDoubleObjectWithLimitsSliderTool(&opengl->near,this);
    near_slider->setOrientation(Qt::Horizontal);
    QToolTip::add(near_slider,"Near Clipping Plane Distance");
    
    far_slider = new XOrsaDoubleObjectWithLimitsSliderTool(&opengl->far,this);
    far_slider->setOrientation(Qt::Horizontal);
    QToolTip::add(far_slider,"Far Clipping Plane Distance");
    
    // or...
    
    ortho_z_near_slider = new XOrsaDoubleObjectWithLimitsSliderTool(&opengl->ortho_z_near_scale,this);
    ortho_z_near_slider->setOrientation(Qt::Horizontal);
    QToolTip::add(ortho_z_near_slider,"Depth Clipping Plane Near Distance");
    
    ortho_z_far_slider = new XOrsaDoubleObjectWithLimitsSliderTool(&opengl->ortho_z_far_scale,this);
    ortho_z_far_slider->setOrientation(Qt::Horizontal);
    QToolTip::add(ortho_z_far_slider,"Depth Clipping Plane Far Distance");
    
    slot_projection_changed();
  }
  
 private:
  OpenGL_Projection projection() const {
    OpenGL_Projection p;
    const QString projection_combo_text = projection_cb->currentText();
    if (projection_combo_text=="Orthographic") {
      p = OGL_ORTHO;
    } else if (projection_combo_text=="Perspective") {
      p = OGL_PERSPECTIVE;
    } else {
      char err_msg[4096];
      sprintf(err_msg,"XOrsaCameraToolBar::projection(): can't handle [%s] projection",projection_combo_text.latin1());
      ORSA_ERROR_MACRO(err_msg);
      //
      p = OGL_ORTHO; // return something...
    }
    return p;
  }
  
 private slots:
  void slot_projection_changed() {
    const OpenGL_Projection p = projection();
    opengl->projection = p;
    widgets_enabler();
  }
  
  void slot_evolution_changed() {
    const orsa::Frame * bodies = opengl->bodies();
    center_cb->Set(   bodies, false);
    eye_cb->Set(      bodies, false);
    rotation_cb->Set( bodies, false);
  }
  
 private slots:  
  void widgets_enabler() {
    
    eye_cb->setEnabled(eye_tb->isOn());
    
    rotation_cb->setEnabled(rotation_tb->isOn());
    
    switch (projection()) {
    case OGL_ORTHO:
      distance_slider->hide();
      //
      eye_tb->setOn(false);
      //
      eye_tb->hide();
      eye_cb->hide();
      //
      FOV_slider->hide();
      near_slider->hide();
      far_slider->hide();
      //
      ortho_xy_slider->show();
      ortho_z_near_slider->show();
      ortho_z_far_slider->show();
      break;
    case OGL_PERSPECTIVE:
      distance_slider->show();
      distance_slider->setEnabled(!eye_tb->isOn());
      //
      eye_tb->show();
      eye_cb->show();
      //
      FOV_slider->show();
      near_slider->show();
      far_slider->show();
      //
      ortho_xy_slider->hide();
      ortho_z_near_slider->hide();
      ortho_z_far_slider->hide();
      break;
    } 
  } 
  
 private:
  XOrsaOpenGLEvolutionWidget * const opengl; 
 private:
  XOrsaImprovedObjectsComboTool * center_cb;
 private:
  XOrsaBoolToolButton * eye_tb;
  XOrsaImprovedObjectsComboTool * eye_cb;
 private:
  XOrsaBoolToolButton * rotation_tb;
  XOrsaImprovedObjectsComboTool * rotation_cb;
 private:
  QComboBox * projection_cb;
 private:
  XOrsaDoubleObjectWithLimitsSliderTool * distance_slider;
  //
  XOrsaDoubleObjectWithLimitsSliderTool *  FOV_slider;
  XOrsaDoubleObjectWithLimitsSliderTool * near_slider;
  XOrsaDoubleObjectWithLimitsSliderTool *  far_slider;
  //
  XOrsaDoubleObjectWithLimitsSliderTool * ortho_xy_slider;
  XOrsaDoubleObjectWithLimitsSliderTool * ortho_z_near_slider;
  XOrsaDoubleObjectWithLimitsSliderTool * ortho_z_far_slider;
};



// the content of the XOrsaExtendedDockOpenGL sidebar
class XOrsaOpenGLSidebarWidget : public QTabWidget {
  
  Q_OBJECT
  
 public:
  XOrsaOpenGLSidebarWidget(XOrsaOpenGL*, QWidget *parent=0);
  
 protected slots:
  void content_update();
  void slot_acquire_single_PS_image();
  void slot_acquire_single_EPS_image();
  void slot_acquire_single_PDF_image();
  void slot_acquire_single_PNG_image();
  void slot_acquire_movie();
  //
  // void set_orthographic_projection() { opengl->SetProjection(OGL_ORTHO); };
  // void set_perspective_projection()  { opengl->SetProjection(OGL_PERSPECTIVE); };
  void set_projection(const QString&);
  //
  void projection_changed_slot();
  //
  void full_screen_slot();
  
 private slots:
  void distance_slider_pressed();
  void distance_slider_changed(int);
  //
  void slot_center_changed();
  void slot_direction_changed();
  //
  void slot_rb_object_toggled(bool);
  //
  void UpdateLocation();
  
 private:
  XOrsaOpenGL *opengl;
  QGroupBox *observer_gb;
  QRadioButton *rb_free, *rb_object;
  XOrsaLocationPushButton *loc_pb;
  
 private:
  QCheckBox *cb_direction, *cb_show_grid;
  XOrsaImprovedObjectsCombo *eoc_center, *eoc_direction;
  QCheckBox *cb_orbits;
  // XOrsaAutoObjectsCombo *oc_orbits;
  XOrsaImprovedObjectsCombo *oc_orbits;
  QCheckBox *cb_moid;
  // XOrsaObjectsCombo *oc_moid_1, *oc_moid_2;
  // XOrsaObjectsCombo *oc_lagrange_1, *oc_lagrange_2;
  XOrsaImprovedObjectsCombo *oc_moid_1, *oc_moid_2;
  XOrsaImprovedObjectsCombo *oc_lagrange_1, *oc_lagrange_2;
  QLabel *center_label;
  bool distance_slider_bool;
  QSlider *distance_slider;
};




class XOrsaExtendedDockOpenGL : public QMainWindow {
  
  Q_OBJECT
  
 public:
  XOrsaExtendedDockOpenGL(QWidget *parent=0);
  XOrsaExtendedDockOpenGL(int w, int h, QWidget *parent=0);
  
  ~XOrsaExtendedDockOpenGL();
  
 private:
  void init_bottom(QStatusBar*);
  void init_dock();
  
 public:
  XOrsaOpenGL *opengl;
  
 private slots:
  void setTimeSliderMaxValue(int);
  void update_zoom_scale_zero();
  void fov_slider_pressed();
  void fov_slider_changed(int);
  void update_fov_slider(int);
  void time_slider_pressed();
  void time_slider_released();
  void play_pb_toggled(bool);
  void projection_changed_slot();
  
 private:
  void keyPressEvent(QKeyEvent*);
  
 private:
  // bool fov_slider_bool, distance_slider_bool;
  bool fov_slider_bool;
  // QSlider *time_slider, *distance_slider, *fov_slider;
  QSlider *time_slider, *fov_slider;
  QPushButton *play_pb;
  bool time_slider_old_animation;
  QDockWindow *dock;
  XOrsaOpenGLSidebarWidget *sidebar;
  // QLabel *fov_label, *distance_label;
  QLabel *fov_label;
  QPushButton *xy_pb, *xz_pb, *yz_pb;
};

#endif // _XORSA_OPENGL_H_
